/* See COPYRIGHT for copyright information. */
#include <u.h>
#include <console.h>

/*
 * Put a number(base <= 16) in a buffer in reverse order; return an
 * optional length and a pointer to the null terminated buffer.
 */
static char *
ksprintn(uvlong uq, int base, int *lenp)
{
	register char *p;
	static char buf[sizeof(uvlong)*8+1];

	p=buf;
	do{
		*++p = "0123456789abcdef"[uq % base];
	}while(uq /= base);
	if(lenp)
		*lenp = p - buf;
	return p;
}

/*
 * Scaled down version of printf(3), but with 'u' as an adverb rather
 * than a verb and 'x' signed, as in Plan 9 print.
 *
 * Space or zero padding and a field width are supported for the numeric
 * formats only. 
 */
static uvlong
getint(va_list *ap, int lflag)
{
	uvlong r;

	switch(lflag){
	case 0:
		r = va_arg(*ap, uint);
		break;

	case 1:
		r = va_arg(*ap, ulong);
		break;

	default:
		r = va_arg(*ap, uvlong);
		break;
	}
	return r;
}

void
kprintf(const char *fmt, va_list ap)
{
	char *p;
	char *q;
	char padc;
	uvlong uq;
	register int ch, n;
	int base, lflag, uflag, tmp, width;

	for(;;){
		padc = ' ';
		width = 0;
		while((ch = *(uchar*)fmt++) != '%'){
			if(ch == '\0')
				return;
			cons_putc(ch);
		}
		lflag = 0;
		uflag = 0;

	Reswitch:
		ch = *(uchar*)fmt++;
		switch(ch){
		case '0':
			padc = '0';
			goto Reswitch;

		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			for(width=0;; ++fmt){
				width = width * 10 + ch - '0';
				ch = *fmt;
				if(ch < '0' || ch > '9')
					break;
			}
			goto Reswitch;

		case 'l':
			lflag++;
			goto Reswitch;

		case 'q':
			lflag += 2;
			goto Reswitch;

		case 'u':
			uflag = 1;
			goto Reswitch;

		case 'b':
			uq = va_arg(ap, int);
			p = va_arg(ap, char *);
			for(q=ksprintn(uq, *p++, nil); (ch = *q--) != '\0';)
				cons_putc(ch);
			if(!uq)
				break;
			for(tmp=0; (n = *p++) != '\0'; ){
				if(uq & (1 << (n - 1))) {
					cons_putc(tmp ? ',' : '<');
					for(; (n = *p) > ' '; ++p)
						cons_putc(n);
					tmp = 1;
				}else
					for(; *p > ' '; ++p)
						continue;
			}
			if(tmp)
				cons_putc('>');
			break;

		case 'c':
			cons_putc(va_arg(ap, int));
			break;

		case 's':
			if((p = va_arg(ap, char*)) == nil)
				p = "(null)";
			while((ch = *p++) != '\0')
				cons_putc(ch);
			break;

		case 'd':
			uq = getint(&ap, lflag);
			if(!uflag && (vlong) uq < 0){
				cons_putc('-');
				uq = -(vlong) uq;
			}
			base = 10;
			goto Number;

		case 'o':
			uq = getint(&ap, lflag);
			base = 8;
			goto Number;

		case 'p':
			cons_putc('0');
			cons_putc('x');
			uq = (ulong)va_arg(ap, void *);
			base = 16;
			goto Number;

		case 'x':
			uq = getint(&ap, lflag);
			if(!uflag && (vlong) uq < 0){
				cons_putc('-');
				uq = -(vlong) uq;
			}
			base = 16;
			/* fall through */
		Number:
			p = ksprintn(uq, base, &tmp);
			if(width && (width -= tmp) > 0)
				while(width--)
					cons_putc(padc);
			for(; tmp>0; tmp--)
				cons_putc(*p--);
			break;

		case '%':
			cons_putc(ch);
			break;

		default:
			cons_putc('%');
			cons_putc(ch);	/* unrecognized flags lost */
			break;
		}
	}
	return;
}

/* like panic, but don't */
void
warn(const char *fmt,...)
{
	va_list ap;

	va_start(ap, fmt);
	kprintf(fmt, ap);
	printf("\n");
	va_end(ap);
}

int
printf(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	kprintf(fmt, ap);
	va_end(ap);
	return 0;
}
